"""Unifi implementation."""

from aiohttp import client_exceptions

from .clients import Clients, URL as client_url
from .devices import Devices, URL as device_url
from .errors import raise_error, ResponseError, RequestError


class Controller:
    """Control a UniFi controller."""

    def __init__(self, host, websession, *,
                 username, password, port=8443, site='default'):
        self.host = host
        self.session = websession
        self.port = port
        self.username = username
        self.password = password
        self.site = site

        self.clients = None
        self.devices = None

    async def login(self):
        url = 'login'
        auth = {
            'username': self.username,
            'password': self.password,
            'remember': True,
        }
        await self.request('post', url, json=auth)

    async def sites(self):
        url = 'self/sites'
        sites = await self.request('get', url)
        return {site['desc']: site for site in sites}

    async def initialize(self):
        clients = await self.request('get', client_url)
        self.clients = Clients(clients, self.request)
        devices = await self.request('get', device_url)
        self.devices = Devices(devices, self.request)

    async def request(self, method, path, json=None):
        """Make a request to the API."""
        url = 'https://{}:{}/api/'.format(self.host, self.port)
        url += path.format(site=self.site)

        try:
            async with self.session.request(method, url, json=json) as res:
                if res.content_type != 'application/json':
                    raise ResponseError(
                        'Invalid content type: {}'.format(res.content_type))
                response = await res.json()
                _raise_on_error(response)
                return response['data']

        except client_exceptions.ClientError as err:
            raise RequestError(
                'Error requesting data from {}: {}'.format(self.host, err)
            ) from None


def _raise_on_error(data):
    """Check response for error message."""
    if isinstance(data, dict) and data['meta']['rc'] == 'error':
        raise_error(data['meta']['msg'])



"""Clients are devices on a UniFi network."""

from .api import APIItems

URL = 's/{site}/stat/sta'


class Clients(APIItems):
    """Represents client network devices."""

    def __init__(self, raw, request):
        super().__init__(raw, request, URL, Client)


class Client:
    """Represents a client network device."""

    def __init__(self, raw, request):
        self.raw = raw
        self._request = request

    @property
    def hostname(self):
        return self.raw.get('hostname')

    @property
    def ip(self):
        return self.raw.get('ip')

    @property
    def is_wired(self):
        return self.raw.get('is_wired')

    @property
    def mac(self):
        return self.raw.get('mac')

    @property
    def name(self):
        return self.raw.get('name')

    @property
    def oui(self):
        return self.raw.get('oui')

    @property
    def site_id(self):
        return self.raw.get('site_id')

    @property
    def sw_mac(self):
        """MAC for switch client is connected to."""
        return self.raw.get('sw_mac')

    @property
    def sw_port(self):
        """Switch port client is connected to."""
        return self.raw.get('sw_port')

    @property
    def wired_rx_bytes(self):
        """Bytes received over wired connection."""
        return self.raw.get('wired-rx_bytes', 0)

    @property
    def wired_tx_bytes(self):
        """Bytes transferred over wired connection."""
        return self.raw.get('wired-tx_bytes', 0)

    def __repr__(self):
        """Return the representation."""
        name = self.name
        if not name:
            name = self.hostname
        return "<Client {}: {} {}\n>".format(name, self.mac, self.raw)



"""Aiounifi errors."""


class AiounifiException(Exception):
    """Base error for aiounifi."""


class RequestError(AiounifiException):
    """Unable to fulfill request.
    Raised when host or API cannot be reached.
    """


class ResponseError(AiounifiException):
    """Invalid response."""


class Unauthorized(AiounifiException):
    """Username is not authorized."""


class LoginRequired(AiounifiException):
    """User is logged out."""


class NoPermission(AiounifiException):
    """Users permissions are read only."""


ERRORS = {
    'api.err.LoginRequired': LoginRequired,
    'api.err.Invalid': Unauthorized,
    'api.err.NoPermission': NoPermission
}


def raise_error(error):
    type = error
    cls = ERRORS.get(type, AiounifiException)
    raise cls("{}".format(type))
